from typing import Any
from lightning import Trainer
from lightning.pytorch.callbacks import (
    OnExceptionCheckpoint as _OnExceptionCheckpoint,
)
from xlm.utils.rank_zero import RankedLogger

logger = RankedLogger(__name__, rank_zero_only=True)


class OnExceptionCheckpoint(_OnExceptionCheckpoint):
    """Same as the base class, but skips saving when exceptions are raised during sanity checking."""

    def on_exception(self, trainer: Trainer, *_: Any, **__: Any) -> None:
        if trainer.sanity_checking:
            return
        # don't save checkpoint if the training has not started yet
        if trainer.global_step <= 1:
            return
        logger.info(
            f"Saving checkpoint on exception at {self.ckpt_path} for epoch {trainer.current_epoch} and global step {trainer.global_step}"
        )
        return super().on_exception(trainer, *_, **__)



import os
from pathlib import Path
from pprint import pformat
from typing import (
    Callable,
    Dict,
    List,
    Literal,
    Protocol,
    Sequence,
    Union,
    TypedDict,
    Mapping,
    Any,
    Optional,
    cast,
)
import shutil
from torchdata.stateful_dataloader import StatefulDataLoader
from torchdata.stateful_dataloader.sampler import (
    RandomSampler,
    StatefulDistributedSampler,
)
from jaxtyping import Integer
from torch import Tensor as TT
from torch.utils.data import DataLoader, IterableDataset, SequentialSampler
from datasets.distributed import split_dataset_by_node
import torch
from transformers import (
    BatchEncoding,
    PreTrainedTokenizer,
    BertTokenizer as _BertTokenizer,
    BertTokenizerFast as _BertTokenizerFast,
    GPT2Tokenizer as _GPT2Tokenizer,
    GPT2TokenizerFast as _GPT2TokenizerFast,
    PreTrainedTokenizerBase,
)
from lightning.pytorch.strategies import DDPStrategy as LDDPStrategy
from lightning.pytorch.strategies import (
    SingleDeviceStrategy as LSingleDeviceStrategy,
)
from lightning.fabric.strategies import DDPStrategy as FDDPStrategy
from lightning.fabric.strategies import (
    SingleDeviceStrategy as FSingleDeviceStrategy,
)
import lightning as L
from xlm import flags
from xlm.utils.imports import get_function
from xlm.utils.rank_zero import rank_zero_only
from xlm.noise import NoiseSchedule
from xlm.utils.rank_zero import RankedLogger
import datasets
from transformers import (
    PreTrainedTokenizer,
    PreTrainedTokenizerBase,
    AddedToken,
)

logger = RankedLogger(__name__, rank_zero_only=True)
ranked_logger = RankedLogger(__name__, rank_zero_only=False)

################################################################################
# region: Types

SingleDeviceStrategy = (LSingleDeviceStrategy, FSingleDeviceStrategy)
DDPStrategy = (LDDPStrategy, FDDPStrategy)


class Tokenizer(Protocol):
    mask_token_id: int
    pad_token_id: int
    cls_token_id: int
    eos_token_id: int
    bos_token_id: int
    mask_token: str
    pad_token: str
    cls_token: str
    eos_token: str
    bos_token: str

    def __call__(
        self,
        text: Union[str, List[str]],
        add_special_tokens: bool = True,
        return_tensors: Optional[str] = None,
    ) -> BatchEncoding: ...

    @property
    def vocab_size(self) -> int: ...

    def __len__(self) -> int: ...

    def decode(
        self,
        token_ids: Union[List[int], Integer[TT, " seq_len"]],
        skip_special_tokens: bool = True,
    ) -> str: ...

    def batch_decode(
        self,
        token_ids: Union[List[List[int]], Integer[TT, " batch seq_len"]],
        skip_special_tokens: bool = True,
    ) -> List[str]: ...

    def convert_tokens_to_ids(
        self, tokens: Union[str, List[str]]
    ) -> Union[int, List[int]]: ...

    def convert_ids_to_tokens(
        self, ids: Union[int, List[int]]
    ) -> Union[str, List[str]]: ...


class BaseBatch(TypedDict):
    """Dict with the keys that are present in input batches for all models.

    Attributes:
        input_ids (Integer[TT, " batch seq_len"]): The input ids to the model.
        attention_mask (Integer[TT, " batch seq_len"]): 1 for tokens that are not padding.
        token_type_ids (Integer[TT, " batch seq_len"]): Can depend on the model type.
            For ILM and IDLM: 0 for CLS, 1 for BOS and prefix, 2 for other tokens.
    """

    input_ids: Integer[TT, " batch seq_len"]
    attention_mask: Integer[TT, " batch seq_len"]
    token_type_ids: Integer[
        TT, " batch seq_len"
    ]  # CLS=0, BOS+REST>=1, prefix=1, non_prefix=2
    # CLEANUP: There is no use for token_type_ids for ARLM and MLM. Instead we should only have constraint tensor.


class DataLoaderKwargs(TypedDict):
    batch_size: int
    num_workers: int
    shuffle: bool
    pin_memory: bool


class Collator(Protocol):
    tokenizer: Tokenizer
    block_size: int
    noise_schedule: NoiseSchedule

    def __call__(
        self,
        batch: List[Mapping[str, Any]],
    ) -> Mapping[str, Any]: ...


# endregion: Types
################################################################################


################################################################################
# region: Tokenizers


class TokenizerMixin:
    """
    1. Adds a `full_vocab_size` property.
    2. provides a `post_creation` method that should be called after creating the tokenizer.
      to ensure all the special tokens are present.
    """

    @property
    def full_vocab_size(self) -> int:
        return self.__len__()

    def post_creation(self):
        """Check the presence of the special tokens and update the post processor."""
        for special_token in [
            "eos_token",
            "bos_token",
            "cls_token",
            "pad_token",
            "mask_token",
            "sep_token",
            "unk_token",
        ]:
            if (token := getattr(self, special_token)) is None:
                raise ValueError(f"{special_token} is not set")


class BertTokenizer(TokenizerMixin, _BertTokenizer):  # type: ignore
    def __init__(self, *args, **kwargs):  # type: ignore
        super().__init__(*args, **kwargs)  # type: ignore
        self.add_special_tokens(
            {"cls_token": "[CLS]", "bos_token": "[BOS]", "eos_token": "[EOS]"}
        )

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: str, *args, **kwargs
    ):
        tokenizer = super().from_pretrained(
            pretrained_model_name_or_path, *args, **kwargs
        )
        tokenizer.post_creation()
        return tokenizer


class BertTokenizerFast(TokenizerMixin, _BertTokenizerFast):  # type: ignore
    def __init__(self, *args, **kwargs):  # type: ignore
        super().__init__(*args, **kwargs)  # type: ignore
        self.add_special_tokens(
            {"cls_token": "[CLS]", "bos_token": "[BOS]", "eos_token": "[EOS]"}
        )

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: str, *args, **kwargs
    ):
        tokenizer = super().from_pretrained(
            pretrained_model_name_or_path, *args, **kwargs
        )
        tokenizer.post_creation()
        return tokenizer


class GPT2Tokenizer(TokenizerMixin, _GPT2Tokenizer):  # type: ignore
    def __init__(self, *args, **kwargs):  # type: ignore
        super().__init__(*args, **kwargs)  # type: ignore
        self.add_special_tokens(
            {
                "cls_token": "<|cls|>",
                "bos_token": "<|bos|>",
                "pad_token": "<|pad|>",
                "unk_token": "<|unk|>",
                "mask_token": "<|mask|>",
                "sep_token": "<|sep|>",
                "eos_token": "<|endoftext|>",  # original
            }
        )

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: str, *args, **kwargs
    ):
        tokenizer = super().from_pretrained(
            pretrained_model_name_or_path, *args, **kwargs
        )
        tokenizer.post_creation()
        return tokenizer


class GPT2TokenizerFast(TokenizerMixin, _GPT2TokenizerFast):  # type: ignore
    def __init__(self, *args, **kwargs):  # type: ignore
        super().__init__(*args, **kwargs)  # type: ignore
        self.add_special_tokens(
            {
                "cls_token": "<|cls|>",
                "bos_token": "<|bos|>",
                "pad_token": "<|pad|>",
                "unk_token": "<|unk|>",
                "mask_token": "<|mask|>",
                "sep_token": "<|sep|>",
                "eos_token": "<|endoftext|>",  # original
            }
        )

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: str, *args, **kwargs
    ):
        tokenizer = super().from_pretrained(
            pretrained_model_name_or_path, *args, **kwargs
        )
        tokenizer.post_creation()
        return tokenizer


class SimpleSpaceTokenizer(PreTrainedTokenizer):
    """Splits on spaces"""

    def __init__(self, vocab: Sequence[str], **kwargs):
        """
        Args:
            vocab (Sequence[str]): List of desired tokens. Following are list of all of the special tokens with
                their corresponding ids:
                    "[PAD]": 0,
                    "[UNK]": 1,
                    "[MASK]": 2,
                    "[EOS]": 3,
                    "[BOS]": 4,
                an id (starting at 5) will be assigned to each character.

            model_max_length (int): Model maximum sequence length.
        """
        self.vocab = vocab
        eos_token = AddedToken("[EOS]", lstrip=False, rstrip=False)
        cls_token = AddedToken("[CLS]", lstrip=False, rstrip=False)
        pad_token = AddedToken("[PAD]", lstrip=False, rstrip=False)
        bos_token = AddedToken("[BOS]", lstrip=False, rstrip=False)
        mask_token = AddedToken("[MASK]", lstrip=True, rstrip=False)
        sep_token = AddedToken("[SEP]", lstrip=False, rstrip=False)
        unk_token = AddedToken("[UNK]", lstrip=False, rstrip=False)

        self._vocab_str_to_int = {
            "[PAD]": 0,
            "[CLS]": 1,
            "[MASK]": 2,
            "[EOS]": 3,
            "[BOS]": 4,
            "[SEP]": 5,
            "[UNK]": 6,
            **{ch: i + 7 for i, ch in enumerate(vocab)},
        }
        self._vocab_int_to_str = {
            v: k for k, v in self._vocab_str_to_int.items()
        }

        super().__init__(
            eos_token=eos_token,
            cls_token=cls_token,
            pad_token=pad_token,
            mask_token=mask_token,
            bos_token=bos_token,
            sep_token=sep_token,
            unk_token=unk_token,
            add_prefix_space=False,
            **kwargs,
        )

    @property
    def vocab_size(self) -> int:
        return len(self._vocab_str_to_int)

    @property
    def full_vocab_size(self) -> int:
        return len(self._vocab_str_to_int)

    def tokenize(self, text: str, **kwargs) -> List[str]:
        # suppose text is a like "split1 split2 split3", convert to character if split* not in vocab
        tokens = []
        for token in text.split(" "):
            if not token:  # skip empty tokens
                continue
            if token not in self._vocab_str_to_int:
                raise ValueError(f"Token {token} not in vocab")
            tokens.append(token)
        return tokens

    def _convert_token_to_id(self, token: str) -> int:
        return self._vocab_str_to_int[token]  # let it raise keyerror

    def _convert_id_to_token(self, index: int) -> str:
        return self._vocab_int_to_str[index]

    def _decode(
        self,
        token_ids: Union[int, List[int]],
        skip_special_tokens: bool = False,
        **kwargs,
    ) -> str:
        return " ".join(
            [self._vocab_int_to_str[token_id] for token_id in token_ids]
        )

    def convert_tokens_to_string(self, tokens):
        return " ".join(tokens)

    def get_vocab(self) -> Dict[str, int]:
        return self._vocab_str_to_int

    @classmethod
    def for_numbers(cls, vocab_size: int, **kwargs) -> "SimpleSpaceTokenizer":
        return cls(vocab=list(map(str, range(vocab_size))), **kwargs)

    def save_pretrained(
        self, save_directory: Union[str, os.PathLike], **kwargs
    ):
        raise NotImplementedError

    @classmethod
    def from_pretrained(
        cls, save_directory: Union[str, os.PathLike], **kwargs
    ):
        raise NotImplementedError

    def from_txt(cls, txt_file: Union[str, os.PathLike], **kwargs):
        vocab = []
        with open(txt_file, "r") as f:
            for line in f:
                vocab.append(line.strip())
        return cls(vocab=vocab, **kwargs)


# endregion: Tokenizers
################################################################################

################################################################################
# region: Preprocess functions


def tokenizing_preprocess_fn(
    example: Dict[str, Any], tokenizer: PreTrainedTokenizerBase
) -> Dict[str, Any]:
    example["token_ids"] = tokenizer.encode(  # type: ignore
        example["text"], add_special_tokens=False
    )
    return example


# endregion: Preprocess functions
################################################################################


################################################################################
# region: Base DataModule


class DatasetManager:
    def __init__(
        self,
        collator: Collator,  # can depend on the model
        # full_name: str,  # e.g. "repo/ds_name/<split>/<type>", eg. `/lm1b/train/lm`
        full_name: str,  # e.g. "repo/ds_name/split", eg. `one-billion-word-benchmark/lm1b/train`
        full_name_debug: str,  # Used with DEBUG_OVERFIT is set to True. So if this manager is for val or test split, replace the split with train here..
        # split_to_download: str,  # e.g. "train", "val", "test", "predict"
        dataloader_kwargs: DataLoaderKwargs,  # e.g. {"batch_size": 1, "num_workers": 1, "shuffle": True, "pin_memory": True}
        preprocess_function: Optional[str] = None,  # e.g. "preprocess_fn"
        on_the_fly_processor: Optional[str] = None,  # e.g. "ids_to_example_fn"
        on_the_fly_group_processor: Optional[
            str
        ] = None,  # e.g. "group_texts_ilm"
        model_name: Optional[
            str
        ] = None,  # used to create model specific cache dir if supplied
        columns_to_remove: Optional[List[str]] = None,  # e.g. ["text"]
        stages: List[
            Literal["fit", "validate", "test", "predict"]
        ] = None,  # e.g. ["fit", "validate"]
        iterable_dataset_shards: Optional[int] = None,  # e.g. 1000
        shuffle_buffer_size: Optional[int] = None,  # e.g. 10000
        shuffle_seed: Optional[int] = 42,  # e.g. 42
        split_by_node: bool = True,  # e.g. True
        dataset: Optional[datasets.Dataset] = None,  # only set later
        rewrite_manual_cache: bool = False,
    ):
        self.collator = collator
        self._full_name = full_name
        self._full_name_debug = full_name_debug
        if flags.DEBUG_OVERFIT:
            logger.info(
                f"Using {self._full_name_debug} instead of {full_name} because DEBUG_OVERFIT is set to True"
            )
            self.full_name = self._full_name_debug
        else:
            self.full_name = full_name
        # self.split_to_download = split_to_download
        self.dataloader_kwargs = dataloader_kwargs
        self.preprocess_function = preprocess_function
        self.on_the_fly_processor = on_the_fly_processor
        self.on_the_fly_group_processor = on_the_fly_group_processor
        self.model_name = model_name
        self.columns_to_remove = columns_to_remove
        self.stages = stages if stages is not None else ["fit", "validate"]
        self.iterable_dataset_shards = iterable_dataset_shards
        self.shuffle_buffer_size = shuffle_buffer_size
        self.shuffle_seed = shuffle_seed
        self.split_by_node = split_by_node
        self.dataset = dataset
        self.dataset = None
        self.is_iterable_dataset = iterable_dataset_shards is not None
        self.rewrite_manual_cache = rewrite_manual_cache

    def __repr__(self) -> str:
        return f"DatasetManager(full_name={self.full_name})"

    @property
    def _split_to_download(self) -> str:
        return self.full_name.split("/")[-1]

    # @property
    # def type(self) -> str:
    #    return self.full_name.split("/")[-1]

    # @property
    # def split(self) -> Literal["train", "val", "test", "predict"]:
    #    return self.full_name.split("/")[-2]

    @property
    def name(self) -> str:
        return "/".join(self.full_name.split("/")[:2])

    def set_epoch(self, epoch: int) -> None:
        # only datasets.IterableDataset has set_epoch to shuffle the buffer
        # TODO (fabric): For map-style datasets that use DistributedSampler, one needs to
        # call set_epoch on the DistributedSampler of the dataloader.
        if hasattr(self.dataset, "set_epoch") and not flags.DEBUG_OVERFIT:
            self.dataset.set_epoch(epoch)  # type: ignore

    def _download(self, num_proc: Optional[int] = None) -> datasets.Dataset:
        name = self.name
        if name[0] == "/":
            name = name[1:]
        logger.info(f"Downloading {self.full_name}")
        ds = datasets.load_dataset(
            name,
            split=self._split_to_download,
            num_proc=num_proc,
            token=os.environ.get("HF_HUB_KEY"),
        )
        return ds

    def _preprocess(
        self,
        ds: datasets.Dataset,
        tokenizer: Tokenizer,
        num_proc: Optional[int] = None,
    ) -> datasets.Dataset:
        if self.preprocess_function is not None:
            preprocess_fn: Callable[[Any, Tokenizer], Any] = get_function(
                self.preprocess_function
            )
            ds = ds.map(
                preprocess_fn,
                batched=False,
                num_proc=num_proc,
                fn_kwargs={"tokenizer": tokenizer},
                remove_columns=self.columns_to_remove,
            )
        return ds

    def _get_cache_dir(self, manual_cache_dir: str) -> Path:
        return Path(manual_cache_dir) / self.full_name

    def _clean_manual_cache(self, manual_cache_dir: str) -> None:
        cache_dir = self._get_cache_dir(manual_cache_dir)
        logger.info(
            f"Cleaning manual cache for {self.full_name} at {cache_dir}"
        )
        # Fallback to pathlib if fsspec fails
        shutil.rmtree(cache_dir, ignore_errors=True)

    def _manually_cache(
        self, ds: datasets.Dataset, manual_cache_dir: str
    ) -> None:
        cache_dir = self._get_cache_dir(manual_cache_dir)
        logger.info(f"Caching {self.full_name} to {cache_dir}")
        ds.save_to_disk(cache_dir, num_shards=self.iterable_dataset_shards)

    def _load_from_cache(self, manual_cache_dir: str) -> datasets.Dataset:
        cache_dir = self._get_cache_dir(manual_cache_dir)
        return datasets.load_from_disk(cache_dir)

    def _check_cache(self, manual_cache_dir: str) -> bool:
        cache_dir = self._get_cache_dir(manual_cache_dir)
        return cache_dir.exists()

    def _apply_on_the_fly_processors(
        self, dataset: datasets.Dataset, tokenizer: Tokenizer
    ):
        if self.on_the_fly_processor is not None:
            processor: Callable = get_function(self.on_the_fly_processor)
            dataset = dataset.map(
                processor, batched=False, fn_kwargs={"tokenizer": tokenizer}
            )
            return dataset
        return dataset

    def _apply_on_the_fly_group_processors(
        self,
        dataset: datasets.Dataset,
        tokenizer: Tokenizer,
        stage: Literal["fit", "validate", "test", "predict"],
        block_size: int,
        pad_id: int,
        type_id_extension: int = 2,
        attn_extension: int = 0,
    ):
        if self.on_the_fly_group_processor is not None:
            group_processor: Callable = get_function(
                self.on_the_fly_group_processor
            )
            dataset = dataset.map(
                group_processor,
                batched=True,
                fn_kwargs={
                    "tokenizer": tokenizer,
                    "block_size": self.block_size,
                    "block_size": block_size,
                    "pad_id": pad_id,
                    "type_id_extension": type_id_extension,
                    "attn_extension": attn_extension,
                },
            )
            return dataset
        return dataset

    def prepare_data(
        self,
        manual_cache_dir: str,
        tokenizer: Tokenizer,
        num_proc: Optional[int] = None,
    ) -> None:
        # check cache, download, preprocess, cache if not already cached.
        if not self._check_cache(manual_cache_dir):
            logger.info(
                f"No manual cache for {self.full_name} found at {manual_cache_dir}"
            )
            ds = self._download(num_proc=num_proc)
            ds = self._preprocess(ds, tokenizer, num_proc=num_proc)
            self._manually_cache(ds, manual_cache_dir)
        elif self.rewrite_manual_cache:
            logger.info(
                f"Rewriting manual cache for {self.full_name} at {manual_cache_dir}"
            )
            ds = self._download(num_proc=num_proc)
            logger.info(
                f"Cleaned {ds.cleanup_cache_files()} automatic cached files"
            )
            self._clean_manual_cache(manual_cache_dir)
            ds = self._preprocess(ds, tokenizer, num_proc=num_proc)
            self._manually_cache(ds, manual_cache_dir)
        else:
            logger.info(
                f"Found manual cache for {self.full_name} at {self._get_cache_dir(manual_cache_dir)}"
            )

    def setup(
        self,
        stage: Literal["fit", "validate", "test", "predict"],
        manual_cache_dir: str,
        tokenizer: Tokenizer,
        block_size: int,
        is_ddp: bool,
        rank: int,
        world_size: int,
    ) -> None:
        if stage in self.stages and self.dataset is None:
            dataset = self._load_from_cache(manual_cache_dir)
            if self.is_iterable_dataset:
                dataset = dataset.to_iterable_dataset(
                    num_shards=self.iterable_dataset_shards
                )
            if (
                self.shuffle_buffer_size is not None
                and not flags.DEBUG_OVERFIT
            ):
                dataset = dataset.shuffle(
                    buffer_size=self.shuffle_buffer_size,
                    seed=self.shuffle_seed,
                )
            dataset = self._apply_on_the_fly_processors(dataset, tokenizer)
            dataset = self._apply_on_the_fly_group_processors(
                dataset,
                tokenizer,
                stage,
                block_size=block_size,
                pad_id=tokenizer.pad_token_id,
                type_id_extension=2,
                attn_extension=0,
            )
            self.dataset = dataset
            if (
                is_ddp
                and world_size > 1
                and self.split_by_node
                and self.is_iterable_dataset
            ):
                logger.info(
                    f"Splitting dataset {self.full_name} by node for {world_size=}"
                )
                self.dataset = split_dataset_by_node(
                    self.dataset, rank=rank, world_size=world_size
                )

    def get_dataloader(
        self,
        type: Literal["train", "val", "test", "predict"],
        is_ddp: bool,
        rank: int,
        world_size: int,
    ) -> Union[DataLoader, StatefulDataLoader]:
        # case 1: DDP and IterDataset=> Dataset is stateful, handles shuffling and already split_by_node. No sampler needed.
        # case 2: DDP and MapDataset=> Need StatefulDistributedSampler with deterministic seed.
        # case 3: Single GPU and MapDataset=> Use StatefulRandomSampler with deterministic seed. Simply setting shuffle=True on StatefulDataloader is equivalent.
        # case 4: Single GPU and IterDataset=> Dataset is stateful and handles shuffling. Does not need splitting.
        if (
            type == "train"
            and is_ddp
            and world_size > 1
            and self.is_iterable_dataset
        ):  # case 1
            DL = StatefulDataLoader
            logger.info(
                f"Using StatefulDataLoader for {type} of {self.full_name} dataloader"
            )
            sampler = None
            # remove shuffle from dataloader kwargs if provided
            if "shuffle" in self.dataloader_kwargs:
                logger.warning(
                    f"Shuffle is set to True for {self.full_name} {type} dataloader for DDP with IterableDataset. This will be ignored as shuffling is handled by the dataset in this case."
                )
                self.dataloader_kwargs.pop("shuffle")
            # Checks:
            num_workers = self.dataloader_kwargs.get("num_workers", 1)
            assert (
                self.iterable_dataset_shards is not None
            ), "iterable_dataset_shards must be set"
            num_shards_per_worker = self.iterable_dataset_shards / (
                world_size * num_workers
            )
            per_worker_batch_size = self.dataloader_kwargs.get("batch_size", 1)
            if num_shards_per_worker > per_worker_batch_size:
                raise ValueError(
                    f"{num_shards_per_worker=} > {per_worker_batch_size=}. This may cause the training to hang even with drop_last=True. "
                    "Either increase the num_workers or world_size to bring down the num_shards_per_worker or reduce the num_shards or increase the batch_size."
                )
        elif (
            type == "train"
            and is_ddp
            and world_size > 1
            and not self.is_iterable_dataset
        ):  # case 2
            # We setup the generator here because we want its seed to be
            # be dependent on the global seed. Letting the sampler create
            # its own generator will use a hardcoded seed of 0 as of today (https://github.com/pytorch/data/issues/1440)
            # while issue 1440 was resolved for RandomSampler, I believe it still exists for StatefulDistributedSampler.
            # Therefore, we continue to send our own seed.
            DL = StatefulDataLoader
            logger.info(
                f"Using StatefulDataLoader for {type} of {self.full_name} dataloader"
            )
            seed = int(torch.empty((), dtype=torch.int64).random_().item())
            sampler = StatefulDistributedSampler(
                self.dataset,
                seed=seed,
                shuffle=True if not flags.DEBUG_OVERFIT else False,
                # num_replicas=world_size, # Let it retrieve these values from the distributed context
                # rank=rank,
            )
        elif (
            type == "train" and not is_ddp and not self.is_iterable_dataset
        ):  # case 3
            DL = StatefulDataLoader
            sampler = (
                RandomSampler(self.dataset)
                if not flags.DEBUG_OVERFIT
                else SequentialSampler(self.dataset)
            )
        elif (
            type == "train" and not is_ddp and self.is_iterable_dataset
        ):  # case 4
            DL = StatefulDataLoader
            sampler = None
        elif type in ["val", "test", "predict"]:
            sampler = None
            DL = DataLoader
            if self.dataloader_kwargs.get("shuffle", False):
                logger.warning(
                    f"Shuffle is set to True for {type} dataloader. This will be ignored as {type} dataloaders are not shuffled."
                )
            self.dataloader_kwargs["shuffle"] = False
        else:
            raise ValueError(f"Invalid dataloader type: {type}")

        dataloader = DL(
            self.dataset,
            collate_fn=self.collator,
            sampler=sampler,
            **self.dataloader_kwargs,
        )
        return dataloader


class LocalDatasetManager(DatasetManager):
    def __init__(
        self,
        *args,
        ds_type: Optional[str] = None,
        load_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        if ds_type is None:
            raise ValueError("ds_type is required")
        super().__init__(*args, **kwargs)
        self.load_kwargs = load_kwargs or {}
        self.ds_type = ds_type

    def _download(self, num_proc: Optional[int] = None) -> datasets.Dataset:
        if self.ds_type == "csv":
            file_name = f"{self._split_to_download}.csv"
            _path = Path(self.full_name).parent
            data_files = str(_path / file_name)
            ds = datasets.load_dataset(
                "csv",
                data_files=data_files,
                **self.load_kwargs,
                num_proc=num_proc,
            )["train"]
            return ds
        else:
            raise ValueError(f"Unsupported dataset type: {self.ds_type}")


class UnconditionalGenerationDatasetManager:
    """
    This is used for unconditional generation, where we don't have any input text.
    """

    def __init__(
        self,
        dataset_constructor_str: str,  # like "xlm.lm.ilm.datamodule_ilm.ILMEmptyDataset"
        collator: Collator,
        num_examples: int,
        dataloader_kwargs: DataLoaderKwargs,
        split_by_node: bool = True,
        dataset_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self.num_examples = num_examples
        self.split_by_node = split_by_node
        self.dataset_constructor: Callable[
            [Tokenizer, int], IterableDataset
        ] = get_function(dataset_constructor_str)
        self.collator = collator
        self.dataloader_kwargs = dataloader_kwargs
        self.dataset_kwargs = dataset_kwargs or {}

    def __repr__(self) -> str:
        return f"ILMUnconditionalGenerationDatasetManager()"

    @property
    def name(self) -> str:
        return "empty"

    def set_epoch(self, epoch: int) -> None:
        return

    def prepare_data(
        self,
        manual_cache_dir: str,
        tokenizer: Tokenizer,
        num_proc: Optional[int] = None,
    ) -> None:
        return

    def setup(
        self,
        stage: Literal["fit", "validate", "test", "predict"],
        manual_cache_dir: str,
        tokenizer: Tokenizer,
        block_size: int,
        is_ddp: bool,
        rank: int,
        world_size: int,
    ) -> None:
        # if is_ddp and world_size > 1 and self.split_by_node:
        #    examples_per_node = self.num_examples // world_size
        # else:
        #    examples_per_node = self.num_examples
        examples_per_node = self.num_examples
        dataset = self.dataset_constructor(
            tokenizer,
            examples_per_node,
            **self.dataset_kwargs,
        )
        self.dataset = dataset

    def get_dataloader(
        self,
        type: Literal["train", "val", "test", "predict"],
        is_ddp: bool,
        rank: int,
        world_size: int,
    ) -> Union[DataLoader, StatefulDataLoader]:
        if type == "train":
            raise ValueError(
                "Train dataloader is not supported for unconditional generation."
            )
        return DataLoader(
            self.dataset, collate_fn=self.collator, **self.dataloader_kwargs
        )


class BaseDataModule(L.LightningDataModule):
    """
    Base class for all datamodules.
    """

    no_trainer_mode: bool = False
    dataloader_names: Dict[
        Literal["train", "val", "test", "predict"], Dict[int, str]
    ]
    dataloader_ids: Dict[
        Literal["train", "val", "test", "predict"], Dict[str, int]
    ]
    tokenizer: Tokenizer
    """The tokenizer.[Required]"""
    global_batch_size: int
    train_dataloader_batch_size: int

    def __init__(self, *args, **kwargs):  # type: ignore
        super().__init__()  # LightningDataModule.__init__() does not take any arguments
        self.prepare_data_per_node = False
        self.no_trainer_mode = False

    def print_batch(
        self,
        batch: BaseBatch,
        split: Literal["train", "val", "test", "predict"],
        dataloader_idx: Optional[int] = None,
    ):
        """Required to print train and validation batches at the beginning of the epoch."""
        raise NotImplementedError("Not implemented")

    def _check_grad_accum(self):
        if self.trainer is None:
            logger.warning("Trainer is not setup. Cannot check grad accum.")
            return
        if self._is_ddp():
            num_nodes = self.trainer.num_nodes
            num_gpus_per_node = self.trainer.num_devices
            accum_steps = self.trainer.accumulate_grad_batches
            logger.info(
                f"global_batch_size: {self.global_batch_size} | num_nodes: {num_nodes} | num_gpus_per_node: {num_gpus_per_node} | accum_steps: {accum_steps}"
            )
            if (
                self.global_batch_size
                % (
                    self.train_dataloader_kwargs["batch_size"]
                    * num_nodes
                    * num_gpus_per_node
                    * accum_steps
                )
                != 0
            ):
                raise ValueError(
                    f"Global batch size ({self.global_batch_size}) is not equal to "
                    f"per_device_batch_size ({self.train_dataloader_kwargs['batch_size']}) * num_nodes ({num_nodes}) * num_gpus_per_node ({num_gpus_per_node}) * accum_steps ({accum_steps})."
                )

    def _is_ddp(self) -> bool:
        if self.trainer is not None:
            strategy = self.trainer.strategy
            if isinstance(strategy, DDPStrategy):
                return True
            elif isinstance(strategy, SingleDeviceStrategy):
                return False
            else:
                raise ValueError(
                    f"Dataloader does not support {type(strategy)} strategy"
                )
        else:
            logger.warning(
                "Tried to detect DDP strategy before trainer was set."
                " Are you calling `LightningDataModule.*_dataloader()` methods manually?"
                " Make sure you know what you are doing!"
            )
            return False

    @property
    def rank(self) -> int:
        if self.trainer is not None:
            return self.trainer.global_rank
        elif self.no_trainer_mode:
            logger.warning(
                "No trainer mode. Returning rank 0. If using multiple GPUs, this will result in incorrect results."
            )
            return 0
        else:
            raise ValueError("Trainer is not set")

    @property
    def world_size(self) -> int:
        if self.trainer is not None:
            return self.trainer.world_size
        elif self.no_trainer_mode:
            logger.warning(
                "No trainer mode. Returning world size 1. If using multiple GPUs, this will result in incorrect results."
            )
            return 1
        else:
            raise ValueError("Trainer is not set")


class TextDataModule(BaseDataModule):
    def __init__(
        self,
        manual_cache_dir: str,
        tokenizer: Tokenizer,
        # dataset_managers: List[DatasetManager],
        dataset_managers: Dict[
            Literal["train", "val", "test", "predict"],
            Dict[str, DatasetManager],  # k=dl_name
        ],
        noise_schedule: Optional[NoiseSchedule] = None,
        rewrite_manual_cache: bool = False,
        block_size: int = 128,
        global_batch_size: int = 512,
        verbosity: Literal["warning", "info", "debug"] = "info",
        num_unconditional_samples: Optional[int] = None,
        num_dataset_workers: Optional[int] = None,
        print_batch_fn: Optional[str] = None,
    ):
        super().__init__()
        self.manual_cache_dir = manual_cache_dir
        self.num_dataset_workers = num_dataset_workers
        self.dataset_managers = dataset_managers
        self.noise_schedule = noise_schedule
        self.block_size = block_size
        self.global_batch_size = global_batch_size
        self.num_unconditional_samples = num_unconditional_samples
        self.tokenizer = tokenizer
        self.rewrite_manual_cache = rewrite_manual_cache
        datasets.utils.logging.set_verbosity(
            datasets.utils.logging.log_levels[verbosity]
        )
        # dataloader names
        self.dataloader_names = {
            "train": {},
            "val": {},
            "test": {},
            "predict": {},
        }
        self.dataloader_ids = {
            "train": {},
            "val": {},
            "test": {},
            "predict": {},
        }
        # check the dataloader_ids for each dataset_manager
        for split, ds_managers in self.dataset_managers.items():
            if split not in ["train", "val", "test", "predict"]:
                raise ValueError(f"Invalid split {split} in dataset_managers")
            if split == "train" and len(ds_managers) != 1:
                raise ValueError(f"There should be exactly one train dataset")
            for dataloader_idx, (
                dataloader_name,
                ds_manager,
            ) in enumerate(ds_managers.items()):
                self.dataloader_names[split][dataloader_idx] = dataloader_name
                self.dataloader_ids[split][dataloader_name] = dataloader_idx

        # determine train dataloader kwargs
        for split, ds_managers in self.dataset_managers.items():
            if split == "train":
                for dataloader_name, ds_manager in ds_managers.items():
                    self.train_dataloader_kwargs = ds_manager.dataloader_kwargs
                    self.train_dataloader_batch_size = (
                        ds_manager.dataloader_kwargs["batch_size"]
                    )

        dl_names_str = pformat(self.dataloader_names)
        logger.info(f"Dataloader names:\n{dl_names_str}")
        if print_batch_fn is not None:
            self.print_batch_fn = get_function(print_batch_fn)
        else:
            self.print_batch_fn = print_batch_base

    def print_batch(
        self,
        batch,
        split: Literal["train", "val", "test", "predict"],
        dataloader_idx: Optional[int] = None,
    ):
        dl_name = (
            self.dataloader_names[split][dataloader_idx]
            if dataloader_idx is not None
            else ""
        )
        self.print_batch_fn(batch, split, self.tokenizer, dl_name)

    def prepare_data(self) -> None:
        ds_wrapper: DatasetManager
        for split, ds_managers in self.dataset_managers.items():
            for dl_name, ds_wrapper in ds_managers.items():
                ds_wrapper.prepare_data(
                    manual_cache_dir=self.manual_cache_dir,
                    tokenizer=self.tokenizer,
                    num_proc=self.num_dataset_workers,
                )

    def setup(
        self, stage: Literal["fit", "validate", "test", "predict"]
    ) -> None:
        ds_wrapper: DatasetManager
        for split, ds_managers in self.dataset_managers.items():
            for dl_name, ds_wrapper in ds_managers.items():
                ds_wrapper.setup(
                    stage,
                    manual_cache_dir=self.manual_cache_dir,
                    tokenizer=self.tokenizer,
                    block_size=self.block_size,
                    is_ddp=self._is_ddp(),
                    rank=self.rank,
                    world_size=self.world_size,
                )

    def _get_dataloaders(
        self, split: Literal["train", "val", "test", "predict"]
    ) -> Any:
        dataloaders = []

        for split_, ds_managers in self.dataset_managers.items():
            for dl_name, ds_wrapper in ds_managers.items():
                if split_ == split:
                    dl = ds_wrapper.get_dataloader(
                        split,
                        is_ddp=self._is_ddp(),
                        rank=self.rank,
                        world_size=self.world_size,
                    )
                    dataloaders.append(dl)

        if split == "train" and len(dataloaders) != 1:
            raise ValueError(
                "there should be exactly one train dataset and dataloader"
            )
        return dataloaders if split != "train" else dataloaders[0]

    def train_dataloader(self) -> Any:
        return self._get_dataloaders("train")

    def val_dataloader(self) -> Any:
        return self._get_dataloaders("val")

    def test_dataloader(self) -> Any:
        return self._get_dataloaders("test")

    def predict_dataloader(self) -> Any:
        return self._get_dataloaders("predict")

    def set_epoch(self, epoch: int) -> None:
        ds_wrapper: DatasetManager
        for split, ds_managers in self.dataset_managers.items():
            for dl_name, ds_wrapper in ds_managers.items():
                ds_wrapper.set_epoch(epoch)


# endregion: Base DataModule
################################################################################


################################################################################
# region: Processors


def token_ids_to_input_ids(
    example: Dict[Literal["token_ids"], List[int]],
    tokenizer: PreTrainedTokenizerBase,
    block_size: Optional[int] = None,
):
    return {
        "input_ids": example["token_ids"][:block_size],
    }


def token_ids_to_input_ids_and_prompt_ids(
    example: Dict[Literal["token_ids"], List[int]],
    tokenizer: PreTrainedTokenizerBase,
    block_size: Optional[int] = None,
):
    return {
        "input_ids": example["input_token_ids"][:block_size],
        "prompt_ids": example["prompt_token_ids"][:block_size],
    }


def ids_to_example_fn(
    example: Dict[Literal["token_ids"], List[int]],
    tokenizer: PreTrainedTokenizerBase,
    block_size: Optional[int] = None,
) -> Dict[Literal["input_ids", "attention_mask", "token_type_ids"], List[int]]:
    """Convert raw token_ids to input_ids, attention_mask, and token_type_ids.

    Does:
        1. Calls `tokenizer.build_inputs_with_special_tokens` and `tokenizer.create_token_type_ids_from_sequences`
            to produce `input_ids` and `token_type_ids`.
        2. Creates an `attention_mask` of all ones.

    Does not do:
        1. Padding/truncation.

    Args:
        example: A dictionary with a "token_ids" key, and value which is a list of token ids.
        tokenizer: A tokenizer that implements `PretrainedTokenizerBase` interface.
            Specifically, it should have `build_inputs_with_special_tokens` and
            `create_token_type_ids_from_sequences` methods overridden if the default
            implementations are not correct.
        block_size: The block size to pad/truncate the input_ids to.
    Returns:
        A dictionary with "input_ids", "attention_mask", and "token_type_ids" keys.
    """
    input_ids = tokenizer.build_inputs_with_special_tokens(
        example["token_ids"]
    )
    attention_mask = [1] * len(input_ids)
    token_type_ids = tokenizer.create_token_type_ids_from_sequences(
        example["token_ids"]
    )
    return {
        "input_ids": input_ids,
        "attention_mask": attention_mask,
        "token_type_ids": token_type_ids,
    }


class DefaultEmptyDataset(IterableDataset):
    def __init__(
        self,
        tokenizer: Tokenizer,
        num_examples: int,
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        empty_text: str = "",
    ):
        """
        Args:
            tokenizer_kwargs: Keyword arguments for the tokenizer.

            empty_text: For MLM, you will want to set the `empty_text` to a sequence of all mask tokens.
        """
        self.tokenizer = tokenizer
        self.num_examples = num_examples
        self.tokenizer_kwargs = tokenizer_kwargs or {}
        self.empty_text = empty_text

    def __iter__(self):
        for _ in range(self.num_examples):
            ex = self.tokenizer(
                self.empty_text,
                add_special_tokens=False,
                **self.tokenizer_kwargs,
            )
            yield ex


# endregion: Processors
################################################################################


################################################################################
# region: Collators


class BaseCollatorInput(TypedDict):
    """Dict with values that are lists of raw input_ids of variable length.

    This is the input to the collator for pre-training.

    The elements of the lists can be of different lengths.

    Attributes:
        input_ids (List[int]): The input ids.
    """

    input_ids: List[int]


class Seq2SeqCollatorInput(TypedDict):
    """Dict with values that are lists of raw input_ids, attention_mask, and token_type_ids.

    This is the input to the collator for pre-training.

    The elements of the lists can be of different lengths.

    Attributes:
        input_ids (List[int]): The input ids.
        prompt_ids (List[int]): The target ids.
    """

    input_ids: List[int]
    prompt_ids: List[int]


class InfillingCollatorInput(TypedDict):
    input_ids: List[int]  # inputs with blanks (masks)
    target_ids: List[int]  # targets with all tokens


class DefaultCollator(Collator):
    """Simply stacks the input_ids, attention_mask, and token_type_ids and returns a batch."""

    def __init__(
        self,
        tokenizer: Tokenizer,
        block_size: int,
        noise_schedule: NoiseSchedule,
    ):
        self.tokenizer = tokenizer
        self.block_size = block_size
        self.noise_schedule = noise_schedule

    @property
    def vocab_size(self) -> int:
        if self.tokenizer is None:
            raise RuntimeError("Tokenizer not set")
        return len(self.tokenizer)

    def __call__(
        self,
        examples: List[BaseCollatorInput],
    ) -> BaseBatch:
        return {
            "input_ids": torch.stack(
                [torch.tensor(e["input_ids"]) for e in examples]
            ),
            "attention_mask": torch.stack(
                [torch.tensor(e["attention_mask"]) for e in examples]
            ),
            "token_type_ids": torch.stack(
                [torch.tensor(e["token_type_ids"]) for e in examples]
            ),
        }


def pad_truncate(
    examples,
    max_len,
    pad_token_id,
    attn_extension: int = 0,
    type_extension: int = 2,
) -> BaseBatch:
    return {
        "input_ids": torch.tensor(
            [
                example["input_ids"][:max_len]
                + [pad_token_id] * max(0, max_len - len(example["input_ids"]))
                for example in examples
            ],
            dtype=torch.long,
        ),
        "attention_mask": torch.tensor(
            [
                example["attention_mask"][:max_len]
                + [attn_extension]
                * max(0, max_len - len(example["attention_mask"]))
                for example in examples
            ],
            dtype=torch.long,
        ),
        "token_type_ids": torch.tensor(
            [
                example["token_type_ids"][:max_len]
                + [type_extension]
                * max(0, max_len - len(example["token_type_ids"]))
                for example in examples
            ],
            dtype=torch.long,
        ),
    }


def pad_prefix_suffix(tokenizer, examples, max_seq_len) -> BaseBatch:
    """
    [<varibale prefix>] [<bos> <variable suffix> <eos>]
    """
    prefixes = []
    suffixes = []
    max_prefix_len = 0
    max_suffix_len = 0
    for example in examples:
        bos_index = example["input_ids"].index(tokenizer.bos_token_id)
        prefix = example["input_ids"][: bos_index + 1]
        max_prefix_len = max(max_prefix_len, len(prefix))
        suffix = example["input_ids"][bos_index + 1 :]
        max_suffix_len = max(max_suffix_len, len(suffix))
        prefixes.append(prefix)
        suffixes.append(suffix)
    return {
        "input_ids": torch.tensor(
            [
                [tokenizer.pad_token_id]
                * max(0, max_prefix_len - len(prefixes[i]))
                + prefixes[i][:max_prefix_len]
                + suffixes[i][:max_suffix_len]
                + [tokenizer.pad_token_id]
                * max(0, max_suffix_len - len(suffixes[i]))
                for i in range(len(examples))
            ][:max_seq_len],
            dtype=torch.long,
        ),
        "attention_mask": torch.tensor(
            [
                [0] * max(0, max_prefix_len - len(prefixes[i]))
                + [1] * len(prefixes[i])
                + [0] * len(suffixes[i])
                + [0] * max(0, max_suffix_len - len(suffixes[i]))
                for i in range(len(examples))
            ][:max_seq_len],
            dtype=torch.long,
        ).bool(),
        "token_type_ids": torch.tensor(
            [
                [2] * max(0, max_prefix_len - len(prefixes[i]))
                + [1] * len(prefixes[i])
                + [2] * len(suffixes[i])
                + [2] * max(0, max_suffix_len - len(suffixes[i]))
                for i in range(len(examples))
            ][:max_seq_len],
            dtype=torch.long,
        ),
    }


def pad_prefix_suffix2(
    examples,
    max_seq_len,
    pad_token_id,
    bos_token_id,
    eos_token_id,
    prefix_type_extension=0,
    suffix_type_extension=2,
    return_tensors=True,
) -> Union[BaseBatch, Dict[str, List[List[int]]]]:
    """
    [<varibale prefix>] [<bos> <variable suffix> <eos>]
    """
    prefixes = []
    suffixes = []
    max_prefix_len = 0
    max_suffix_len = 0
    for example in examples:
        bos_index = example["input_ids"].index(bos_token_id)
        prefix = example["input_ids"][: bos_index + 1]
        max_prefix_len = max(max_prefix_len, len(prefix))
        suffix = example["input_ids"][bos_index + 1 :]
        max_suffix_len = max(max_suffix_len, len(suffix))
        prefixes.append(prefix)
        suffixes.append(suffix)
    input_ids = [
        [pad_token_id] * max(0, max_prefix_len - len(prefixes[i]))
        + prefixes[i][:max_prefix_len]
        + suffixes[i][:max_suffix_len]
        + [pad_token_id] * max(0, max_suffix_len - len(suffixes[i]))
        for i in range(len(examples))
    ]
    attention_mask = [
        [0] * max(0, max_prefix_len - len(prefixes[i]))
        + [1] * len(prefixes[i])
        + [0] * len(suffixes[i])
        + [0] * max(0, max_suffix_len - len(suffixes[i]))
        for i in range(len(examples))
    ]
    token_type_ids = [
        [prefix_type_extension] * max(0, max_prefix_len - len(prefixes[i]))
        + [1] * len(prefixes[i])
        + [suffix_type_extension] * len(suffixes[i])
        + [suffix_type_extension] * max(0, max_suffix_len - len(suffixes[i]))
        for i in range(len(examples))
    ]
    if return_tensors:
        input_ids = torch.tensor(input_ids, dtype=torch.long)  # type: ignore
        attention_mask = torch.tensor(attention_mask, dtype=torch.bool)  # type: ignore
        token_type_ids = torch.tensor(token_type_ids, dtype=torch.long)  # type: ignore

    return {
        "input_ids": input_ids,
        "attention_mask": attention_mask,
        "token_type_ids": token_type_ids,
    }


def pad_left_truncate(
    examples,
    max_len,
    pad_token_id,
    attn_extension: int = 0,
    type_extension: int = 2,
    return_tensors: bool = True,
) -> Union[BaseBatch, Dict[str, List[List[int]]]]:
    input_ids: List[List[int]] = [
        [pad_token_id] * max(0, max_len - len(example["input_ids"]))
        + example["input_ids"][:max_len]
        for example in examples
    ]
    attention_mask: List[List[int]] = [
        [attn_extension] * max(0, max_len - len(example["attention_mask"]))
        + example["attention_mask"][:max_len]
        for example in examples
    ]
    token_type_ids: List[List[int]] = [
        [type_extension] * max(0, max_len - len(example["token_type_ids"]))
        + example["token_type_ids"][:max_len]
        for example in examples
    ]
    if return_tensors:
        input_ids = torch.tensor(input_ids, dtype=torch.long)  # type: ignore
        attention_mask = torch.tensor(attention_mask, dtype=torch.bool)  # type: ignore
        token_type_ids = torch.tensor(token_type_ids, dtype=torch.long)  # type: ignore

    return {
        "input_ids": input_ids,
        "attention_mask": attention_mask,
        "token_type_ids": token_type_ids,
    }


def pad_dynamic(
    examples,
    pad_token_id,
    attn_extension: int = 0,
    type_extension: int = 2,
) -> BaseBatch:
    max_len = max(len(e["input_ids"]) for e in examples)
    return pad_truncate(
        examples,
        max_len,
        pad_token_id,
        attn_extension=attn_extension,
        type_extension=type_extension,
    )


class DefaultCollatorWithPadding(DefaultCollator):
    """Like DefaultCollator, but pads (truncates if needed) the input_ids, attention_mask, and token_type_ids to self.max_length."""

    def get_max_len(self, examples: List[BaseCollatorInput]) -> int:
        return self.block_size

    def __call__(self, examples: List[BaseCollatorInput]) -> BaseBatch:
        max_len = self.get_max_len(examples)
        return pad_truncate(
            examples,
            max_len,
            self.tokenizer.pad_token_id,
            attn_extension=0,
            type_extension=2,
        )


class DefaultCollatorWithDynamicPadding(DefaultCollatorWithPadding):
    """Like DefaultCollator, but pads to the max length in the batch."""

    def get_max_len(self, examples: List[BaseCollatorInput]) -> int:
        max_in_batch = max(len(e["input_ids"]) for e in examples)
        return min(max_in_batch, self.block_size)


# endregion: Collators
################################################################################


################################################################################
# region: Utilities


def print_batch_base(
    batch: Dict[str, Any],
    split: Literal["train", "val", "test", "predict"],
    tokenizer: Tokenizer,
    dataloader_name: str = "",
):

    logger.info(
        f"Printing first entries of the tensors in batch for {split}/{dataloader_name}..."
    )
    print("input tokens:")
    print(tokenizer.decode(batch["input_ids"][0]))
    print("input_ids:")
    print(batch["input_ids"][0])
    print("attention_mask:")
    print(batch["attention_mask"][0])
    print("token_type_ids:")
    print(batch["token_type_ids"][0])


# endregion: Utilities
################################################################################



import re
from typing import Any, Dict

from transformers import PreTrainedTokenizerBase
from xlm.utils.rank_zero import RankedLogger

logger = RankedLogger(__name__, rank_zero_only=True)
ranked_logger = RankedLogger(__name__, rank_zero_only=False)


def detokenizer_lm1b_string(x: str) -> str:
    """Detokenizer for the LM1B dataset.
    Same as the one used in SEDD and MDLM.
    """
    # MAYBE: Fix the issues in the detokenizer. Run it on the `detok_test_set.txt` to see the issues.
    x = x.replace("http : / / ", "http://")
    x = x.replace("https : / / ", "https://")
    x = re.sub(
        r" \'(\w+)", r"'\1", x
    )  # remove extra space before single quotes like "they 've"
    x = re.sub(r" (\w+) \. ", r" \1. ", x)
    x = re.sub(r" (\w+) \.$", r" \1.", x)
    x = x.replace(" ? ", "? ")
    x = re.sub(r" \?$", "?", x)
    x = x.replace(" ! ", "! ")
    x = re.sub(r" \!$", "!", x)
    x = x.replace(" , ", ", ")
    x = x.replace(" : ", ": ")
    x = x.replace(" ; ", "; ")
    x = x.replace(" / ", "/")
    x = re.sub(r"\" ([^\"]+) \"", r'"\1"', x)
    x = re.sub(r"\' ([^\']+) \'", r"'\1'", x)
    x = re.sub(r"\( ([^\(\)]+) \)", r"(\1)", x)
    x = re.sub(r"\[ ([^\[\]]+) \]", r"[\1]", x)
    x = x.replace("$ ", "$")
    x = x.replace("£ ", "£")
    return x


def preprocess_fn(
    example: Dict[str, Any], tokenizer: PreTrainedTokenizerBase
) -> Dict[str, Any]:
    example["text"] = detokenizer_lm1b_string(example["text"])
    example["token_ids"] = tokenizer.encode(  # type: ignore
        example["text"],
        add_special_tokens=False,
    )
    # truncate to 128
    example["token_ids"] = example["token_ids"][:128]
    return example



import logging

logger = logging.getLogger(__name__)


def set_flags(cfg):
    # get all flags under cfg.global_flags if exists
    global_flags = cfg.get("global_flags", None)
    if global_flags is None:
        return
    for flag, value in global_flags.items():
        import xlm.flags as flags

        logger.info(f"Setting flag {flag} to {value}")
        flags.__dict__[flag] = value



# %%
# change dir to the root of the project
# create the notebook inside the commands directory
import os

import torch


if "PROJECT_ROOT" not in os.environ:
    os.environ["PROJECT_ROOT"] = "."
os.environ["HYDRA_FULL_ERROR"] = "1"

# region: Import necessary modules
from pathlib import Path
import hydra
from hydra.core.config_store import ConfigStore
from dataclasses import dataclass
import logging

logger = logging.getLogger(__name__)

# endregion


# Hydra configuration parameters


@dataclass
class ExtractModelStateConfig:
    checkpoint_path: Path
    output_path: Path
    add_step_info: bool = False


cs = ConfigStore.instance()
cs.store(name="extract_model_state_dict", node=ExtractModelStateConfig)

_HYDRA_PARAMS = {
    "version_base": "1.3",
    # "config_path": str(Path("../../../configs") / "lightning_train"),
    "config_name": "extract_model_state_dict.yaml",
}


def print_checkpoint_info(full_state_dict: dict) -> None:
    logger.info(f"Checkpoint keys: {full_state_dict.keys()}")
    for k, v in full_state_dict.items():
        if isinstance(v, (int, float, bool, str)):
            logger.info(f"full_state_dict[{k}]: {v}")


@hydra.main(**_HYDRA_PARAMS)
def main(cfg: ExtractModelStateConfig) -> None:
    with open(cfg.checkpoint_path, "rb") as f:
        _full_state_dict = torch.load(f, weights_only=False)
        full_state_dict = _full_state_dict[
            "state_dict"
        ]  # lightning module state_dict
        print_checkpoint_info(_full_state_dict)
        if cfg.add_step_info:
            step = _full_state_dict.get("global_step", None)
            epoch = _full_state_dict.get("epoch", None)
            prefix = f"{epoch=}-{step=}_"
        else:
            prefix = ""
        model_state_dict = {}
        for key in full_state_dict.keys():
            if key.startswith("model."):
                model_state_dict[key[6:]] = full_state_dict[key]
    cfg.output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path = Path(cfg.output_path)
    fname = prefix + output_path.name
    output_path = output_path.parent / fname
    torch.save(model_state_dict, output_path)
    logger.info(f"Model state dict saved to {output_path}")


if __name__ == "__main__":
    main()



# %%
import os
from pathlib import Path
from typing import Optional
from xlm.tasks.star import (
    SimpleStarGraphGenerator,
    AsymmetricVariableArmlengthStarGraphGenerator,
)
import datasets
from xlm.utils.seed import seed_everything_simple
import dotenv

dotenv.load_dotenv(override=True)
dotenv.load_dotenv(".secrets.env", override=True)
seed_everything_simple(42)

datasets.logging.set_verbosity_info()

# start_small i.e., the "easy" setting
SIZES = {
    "train": 50000,
    "validation": 100,
    "test": 5000,
}

STAR_EASY = {
    "constructor": SimpleStarGraphGenerator,
    "kwargs": {
        "degree": 3,
        "pathlength": 5,
        "vocab_size": 20,
    },
    "name": "star-small",
}

STAR_MEDIUM = {
    "constructor": AsymmetricVariableArmlengthStarGraphGenerator,
    "kwargs": {
        "degree": 3,
        "min_pathlength": 3,
        "max_pathlength": 6,
        "min_plan_length": 2,
        "vocab_size": 20,
    },
    "name": "star-medium",
}

STAR_HARD = {
    "constructor": AsymmetricVariableArmlengthStarGraphGenerator,
    "kwargs": {
        "degree": 5,
        "min_pathlength": 6,
        "max_pathlength": 12,
        "min_plan_length": 5,
        "vocab_size": 56,
    },
    "name": "star-hard",
}


# %%
def dump_dataset(
    settings: dict, path: Optional[Path] = None, push_to_hub: bool = False
):
    name = settings["name"]
    datasets_: dict[str, datasets.Dataset] = {}
    for split, size in SIZES.items():
        generator = settings["constructor"](
            count=size,
            **settings["kwargs"],
        )

        dataset = datasets.Dataset.from_generator(
            generator.graphs,
            keep_in_memory=True,
            num_proc=1,
        )
        datasets_[split] = dataset

    dataset_dict = datasets.DatasetDict(datasets_)
    if push_to_hub:
        dataset_dict.push_to_hub(
            f"dhruveshpatel/{name}", token=os.getenv("HF_HUB_KEY")
        )
    if path is not None:
        dataset_dict.save_to_disk(path)
    return dataset


# %%
if __name__ == "__main__":
    # %%
    dump_dataset(STAR_EASY, path=None, push_to_hub=True)
    dump_dataset(STAR_MEDIUM, path=None, push_to_hub=True)
    dump_dataset(STAR_HARD, path=None, push_to_hub=True)

# %%



from curses import savetty
import logging
from pathlib import Path
from typing import Sequence, Union

import omegaconf
import torch

import rich
import rich.syntax
import rich.tree
from hydra.core.hydra_config import HydraConfig
from omegaconf import DictConfig, OmegaConf, open_dict

from .os import is_notebook
from .rank_zero import rank_zero_only
from rich.prompt import Prompt

logger = logging.getLogger(__name__)


@rank_zero_only
def print_config_tree(
    cfg: DictConfig,
    print_order: Sequence[str] = [],
    resolve: bool = True,
    save_to_file: Union[bool, str] = False,
) -> None:
    """Prints content of DictConfig using Rich library and its tree structure.

    Args:
        cfg (DictConfig): Configuration composed by Hydra.
        print_order (Sequence[str], optional): Determines what top level fields to print and in what order.
        resolve (bool, optional): Whether to resolve reference fields of
            DictConfig.
        save_to_file (bool, optional): Whether to export config to the hydra
            output folder.
    """
    # Saving as yaml file is better for readability
    yaml = omegaconf.OmegaConf.to_yaml(cfg, resolve=resolve)
    # save to file
    if save_to_file:
        if isinstance(save_to_file, str):
            dir_ = Path(save_to_file)
        else:
            dir_ = Path(cfg.paths.output_dir)

        dir_.mkdir(parents=True, exist_ok=True)
        logger.info(f"Saving config tree to {dir_ / 'config_tree.yaml'}")
        with open(dir_ / "config_tree.yaml", "w") as file:
            file.write(yaml)

    # detect notebook and disable rich
    if is_notebook():
        print(yaml)
        return

    style = "dim"
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)

    queue = []

    # add fields from `print_order` to queue

    for field in print_order:
        (
            queue.append(field)
            if field in cfg
            else logger.warning(
                f"Field '{field}' not found in config. Skipping '{field}' config printing..."
            )
        )

    # add all the other fields to queue (not specified in `print_order`)
    for field in cfg:
        if field not in queue:
            queue.append(field)

    # generate config tree from queue
    for field in queue:
        branch = tree.add(field, style=style, guide_style=style)

        config_group = cfg[field]
        if isinstance(config_group, DictConfig):
            branch_content = OmegaConf.to_yaml(config_group, resolve=resolve)
        else:
            branch_content = str(config_group)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    # print config tree
    rich.print(tree)

    # save config tree to file
    # Create the output directory if it doesn't exist
    Path(cfg.paths.output_dir).mkdir(parents=True, exist_ok=True)
    if save_to_file:
        logger.info(
            f"Saving config tree to {Path(cfg.paths.output_dir) / 'config_tree.txt'}"
        )
        with open(Path(cfg.paths.output_dir) / "config_tree.txt", "w") as file:
            console = rich.console.Console(file=file)
            console.print(tree)


@rank_zero_only
def enforce_tags(cfg: DictConfig, save_to_file: bool = False) -> None:
    """Prompts user to input tags from command line if no tags are provided in
    config."""
    if not cfg.get("tags"):
        if "id" in HydraConfig().cfg.hydra.job:
            raise ValueError("Specify tags before launching a multirun!")

        logger.warning(
            "No tags provided in config. Prompting user to input tags..."
        )
        tags = Prompt.ask(
            "Enter a list of comma separated tags", default="dev"
        )
        tags = [t.strip() for t in tags.split(",") if t != ""]

        with open_dict(cfg):
            cfg.tags = tags

        logger.info(f"Tags: {cfg.tags}")

    if save_to_file:
        with open(Path(cfg.paths.output_dir, "tags.log"), "w") as file:
            rich.print(cfg.tags, file=file)



from typing import List
from setuptools import setup, find_packages
import os


VERSION = {}  # type: ignore
with open("src/xlm/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

PATH_ROOT = os.path.dirname(__file__)

def load_requirements(
    path_dir: str = PATH_ROOT, comment_char: str = "#"
) -> List:
    with open(os.path.join(path_dir, "requirements.txt"), "r") as file:
        reqs = [ln.strip() for ln in file.readlines()]
    return reqs

install_requires = load_requirements()

setup(
    name="xlm-core",
    version=VERSION["VERSION"],
    author="Dhruvesh Patel, Benjamin Rozonoyer, Sai Sreenivas Chintha, Durga Prasad Maram",
    packages=find_packages(
        where="src",
        exclude=[
            "*.tests",
            "*.tests.*",
            "tests.*",
            "tests",
        ],
    ),
    description='XLM Framework',
    long_description="""
    XLM is a unified framework for developing and comparing small non-autoregressive language models. It uses PyTorch as the deep learning framework, PyTorch Lightning for training utilities, and Hydra for configuration management. XLM provides core components for flexible data handling and training, useful architectural implementations for non-autoregressive workflows, and support for arbitrary runtime code injection. Custom model implementations that leverage the core components of xlm can be found in the xlm-models package. The package also includes a few preconfigured synthetic planning and language-modeling datasets.

    Usage:
        pip install xlm-core

    Command usage:
        xlm job_type=[JOB_TYPE] job_name=[JOB_NAME] experiment=[CONFIG_PATH]
       
        The job_type argument can be one of train ,eval and generate. The experiment argument should point to the root hydra config file.
""",
    install_requires = install_requires,
    project_urls={
        "Source Code": "https://github.com/dhruvdcoder/xlm-core"
    },
    package_dir={"": "src"},
    package_data={
        "xlm": ["configs/**/*.yaml", "configs/**/*.yml"],
    },
    keywords=[
        "AI",
        "ML",
        "Machine Learning",
        "Deep Learning",
        "Non-Autoregressive Language Models",
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "xlm=xlm.__main__:main",
            "xlm-scaffold=xlm.commands.scaffold_model:main",
            "xlm-push-to-hub=xlm.commands.push_to_hub:main"
        ],
    },
    python_requires=">=3.11",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import os
import datasets
from transformers import GPT2TokenizerFast
import dotenv

dotenv.load_dotenv(override=False)
dotenv.load_dotenv(".secrets.env", override=True)
datasets.utils.logging.set_verbosity(
    datasets.utils.logging.log_levels["debug"]
)


def filter_fn(example, tokenizer):
    token_ids = tokenizer.encode(example["text"], add_special_tokens=False)
    return len(token_ids) <= 1024


def main():
    dataset_name = "Skylion007/openwebtext"
    num_proc = 10
    dataset = datasets.load_dataset(
        dataset_name,
        num_proc=num_proc,
        trust_remote_code=True,
    )

    # tokenizer using GPT2Tokenizer and remove sequences longer than 1024 tokens
    tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")
    print(f"Before filtering:\n {dataset}")

    dataset = dataset.filter(
        filter_fn,
        num_proc=num_proc,
        fn_kwargs={"tokenizer": tokenizer},
        desc="Filtering dataset",
    )
    print(f"After filtering:\n {dataset}")

    # split the dataset into train and val
    # When downloaded, it has only one split "train"
    split_datasets = dataset["train"].train_test_split(
        test_size=10000, seed=2357, shuffle=True
    )
    # rename the test split to validation
    split_datasets["validation"] = split_datasets.pop("test")
    print(f"After splitting:\n {split_datasets}")

    # push to huggingface hub
    split_datasets.push_to_hub(
        "dhruveshpatel/owt-gpt2-1024-split", token=os.getenv("HF_HUB_KEY")
    )


if __name__ == "__main__":
    main()



from contextlib import contextmanager
from typing import Any, List, Literal, Optional, Tuple, Union, overload
import torch


def tiny_value_of_dtype(dtype: torch.dtype):
    """
    Returns a moderately tiny value for a given PyTorch data type that is used to avoid numerical
    issues such as division by zero.
    This is different from `info_value_of_dtype(dtype).tiny` because it causes some NaN bugs.
    Only supports floating point dtypes.
    """
    if not dtype.is_floating_point:
        raise TypeError("Only supports floating point dtypes.")
    if dtype == torch.float or dtype == torch.double:
        return 1e-13
    elif dtype == torch.half:
        return 1e-4
    else:
        raise TypeError("Does not support dtype " + str(dtype))


def masked_mean(
    vector: torch.Tensor,
    mask: torch.BoolTensor,
    dim: int,
    keepdim: bool = False,
) -> torch.Tensor:
    """
    To calculate mean along certain dimensions on masked values

    # Parameters

    vector : `torch.Tensor`
        The vector to calculate mean.
    mask : `torch.BoolTensor`
        The mask of the vector. It must be broadcastable with vector.
        It must be 1 for non-masked values and 0 for masked values.
    dim : `int`
        The dimension to calculate mean
    keepdim : `bool`
        Whether to keep dimension

    # Returns

    `torch.Tensor`
        A `torch.Tensor` of including the mean values.
    """
    replaced_vector = vector.masked_fill(~mask, 0.0)

    value_sum = torch.sum(replaced_vector, dim=dim, keepdim=keepdim)
    dtype = vector.dtype
    value_count = (
        torch.sum(mask, dim=dim, keepdim=keepdim).to(dtype=dtype) + 1e-9
    )
    # tiny_value_of_dtype(dtype)
    return value_sum / value_count


def masked_sum(
    vector: torch.Tensor,
    mask: torch.BoolTensor,
    dim: int,
    keepdim: bool = False,
) -> torch.Tensor:
    """
    To calculate sum along certain dimensions on masked values

    # Parameters

    vector : `torch.Tensor`
        The vector to calculate sum.
    mask : `torch.BoolTensor`
        The mask of the vector. It must be broadcastable with vector.
    dim : `int`
        The dimension to calculate sum
    keepdim : `bool`
        Whether to keep dimension

    # Returns

    `torch.Tensor`
        A `torch.Tensor` of including the sum values.
    """
    replaced_vector = vector.masked_fill(~mask, 0.0)

    return torch.sum(replaced_vector, dim=dim, keepdim=keepdim)


def get_mask_from_sequence_lengths(
    sequence_lengths: torch.Tensor, max_length: int
) -> torch.BoolTensor:
    """
    Given a variable of shape `(batch_size,)` that represents the sequence lengths of each batch
    element, this function returns a `(batch_size, max_length)` mask variable.  For example, if
    our input was `[2, 2, 3]`, with a `max_length` of 4, we'd return
    `[[1, 1, 0, 0], [1, 1, 0, 0], [1, 1, 1, 0]]`.

    We require `max_length` here instead of just computing it from the input `sequence_lengths`
    because it lets us avoid finding the max, then copying that value from the GPU to the CPU so
    that we can use it to construct a new tensor.
    """
    # (batch_size, max_length)
    ones = sequence_lengths.new_ones(sequence_lengths.size(0), max_length)
    range_tensor = ones.cumsum(dim=1)
    return sequence_lengths.unsqueeze(1) >= range_tensor


dtype_map = {
    "float32": torch.float32,
    "float64": torch.float64,
    "float16": torch.float16,
    "bfloat16": torch.bfloat16,
}


def dtype(string: str) -> torch.dtype:
    """
    Convert a string to a PyTorch data type.

    # Parameters

    string : `str`
        The string to convert.

    # Returns

    `torch.dtype`
        The PyTorch data type.
    """
    if string in dtype_map:
        return dtype_map[string]
    else:
        raise ValueError(f"Unknown dtype: {string}")


# Define a function to select the appropriate dtype
def get_autocast_dtype():
    if (
        torch.cuda.get_device_properties(0).major >= 8
    ):  # Ampere and later architectures
        return torch.bfloat16
    else:
        return torch.float16


@contextmanager
def no_grad(no_grad_: bool = True):
    if no_grad_:
        with torch.no_grad():
            yield
    else:
        yield


def sample_categorical(probs: torch.Tensor) -> torch.Tensor:
    """Need this since torch.multinomial does not accept 3D input and cannot handle unnormalized probabilities.

    So we implement the "exponential race method" manually which can handle any number of leading dimensions
    and can handle unnormalized probabilities (not logits, )

    Note: This is not differentiable.. Use gumbel softmax for it.
    Args:
        probs: (*batch, seq_len, vocab_size) can have any number of leading dimensions.
            Note: probs should be positive, can be unnormalized.
    Returns:
        (*batch, seq_len)
    """
    exp_1_samples = 1e-10 - (torch.rand_like(probs) + 1e-10).log()
    return (probs / exp_1_samples).argmax(dim=-1)


def sample_from_logits(
    logits: torch.Tensor, temperature: float = 1.0, noise_scale: float = 1.0
) -> torch.Tensor:
    """
    Sample from logits using the Gumbel-Max trick. Similar to sample_categorical, but works with logits (real valued).
    Args:
        logits: (*batch, seq_len, vocab_size) can have any number of leading dimensions.
    Returns:
        (*batch, seq_len)
    """
    gumbel_noise = -torch.log(
        -torch.log(torch.rand_like(logits) + 1e-10) + 1e-10
    )
    perturbed_logits = (logits / temperature) + gumbel_noise * noise_scale
    return perturbed_logits.argmax(dim=-1)


def sample_from_top_k(k: int, logits: torch.Tensor) -> torch.Tensor:
    """
    Sample from the top-k logits using the Gumbel-Max trick.
    Args:
        logits: (*batch, seq_len, vocab_size) can have any number of leading dimensions.
        k: The number of top logits to consider for sampling.
    Returns:
        (*batch, seq_len)
    """
    # Get the top-k logits and their indices
    if logits.shape[-1] < k:
        k = logits.shape[-1]
    top_k_logits, top_k_indices = torch.topk(
        logits, k, dim=-1
    )  # (*batch, seq_len, k)

    # Add Gumbel noise to the top-k logits
    gumbel_noise = -torch.log(
        -torch.log(torch.rand_like(top_k_logits) + 1e-10) + 1e-10
    )
    perturbed_top_k_logits = (
        top_k_logits + gumbel_noise
    )  # (*batch, seq_len, k)

    # Sample from the perturbed top-k logits
    sampled_indices = perturbed_top_k_logits.argmax(
        dim=-1
    )  # (*batch, seq_len)

    # Map back to the original indices
    sampled_indices_full = top_k_indices.gather(
        -1, sampled_indices.unsqueeze(-1)
    ).squeeze(-1)

    return sampled_indices_full  # (*batch, seq_len)


def sample_from_top_p(p: float, logits: torch.Tensor) -> torch.Tensor:
    """
    Sample from the top-p logits using the Gumbel-Max trick.

    Args:
        p (float): The cumulative probability threshold. Must be between 0 and 1.
        logits (torch.Tensor): A tensor of shape (*batch, seq_len, vocab_size) representing
                               the unnormalized log probabilities for each token.

    Returns:
        torch.Tensor: A tensor of shape (*batch, seq_len) containing the sampled token indices.
    """
    if not (0.0 < p <= 1.0):
        raise ValueError(f"Parameter p must be in (0, 1], got {p}")

    # Compute softmax probabilities
    probs = torch.softmax(
        logits, dim=-1
    )  # Shape: (*batch, seq_len, vocab_size)

    # Sort the probabilities in descending order (highest first)
    sorted_probs, sorted_indices = torch.sort(probs, descending=True, dim=-1)

    # Compute the cumulative sum of the sorted probabilities
    cumulative_probs = torch.cumsum(
        sorted_probs, dim=-1
    )  # Shape: (*batch, seq_len, vocab_size)

    # Create a mask for tokens to keep: cumulative_probs <= p
    # This keeps the top tokens until we reach cumulative probability p
    sorted_indices_to_keep = cumulative_probs <= p

    # Ensure that at least one token is kept (the highest probability one)
    sorted_indices_to_keep[..., 0] = True

    # Initialize a mask for the original logits
    mask = torch.zeros_like(
        logits, dtype=torch.bool
    )  # Shape: (*batch, seq_len, vocab_size)

    # Scatter the sorted mask back to the original logits' positions
    mask.scatter_(-1, sorted_indices, sorted_indices_to_keep)

    # Mask out logits that are not in the top-p subset by setting them to -inf
    masked_logits = logits.masked_fill(~mask, float("-inf"))

    # Add Gumbel noise to the masked logits
    gumbel_noise = -torch.log(
        -torch.log(torch.rand_like(masked_logits) + 1e-10) + 1e-10
    )
    perturbed_logits = masked_logits + gumbel_noise

    # Sample the token with the highest perturbed logit
    sampled_indices = perturbed_logits.argmax(
        dim=-1
    )  # Shape: (*batch, seq_len)

    return sampled_indices  # (*batch, seq_len)


# overload for tensor for typing
@overload
def pad_truncate_list(
    ids: List[Any],
    max_len: int,
    pad_token: Any,
    pad_left: bool = False,
    return_num_padded: bool = False,
) -> List[Any]: ...


@overload
def pad_truncate_list(
    ids: List[Any],
    max_len: int,
    pad_token: Any,
    pad_left: bool = False,
    return_num_padded: bool = True,
) -> Tuple[List[Any], int]: ...


def pad_truncate_list(
    ids: List[Any],
    max_len: int,
    pad_token: Any,
    pad_left: bool = False,
    return_num_padded: bool = False,
) -> Union[List[Any], Tuple[List[Any], int]]:
    num_padded = max_len - len(ids)
    if not pad_left:
        padded = ids[:max_len] + [pad_token] * num_padded
    else:
        padded = [pad_token] * num_padded + ids[
            -max_len:
        ]  # when padding left, truncate left side
    if return_num_padded:
        return padded, num_padded
    else:
        return padded


def add_gumbel_noise(
    logits: torch.Tensor, temperature: float = 1.0, noise_scale: float = 1.0
) -> torch.Tensor:
    """
    Add gumbel noise to logits which will result in samples from the distribtution if argmaxed.
    Args:
        logits: (*batch, seq_len, vocab_size) can have any number of leading dimensions. Assumed to be log of exponentiated scores.
            That is, we assume logits are $l_i$ in $p_i = exp(l_i) / \sum_i \exp(l_i)$.
    Returns:
        (*batch, seq_len)
    """
    gumbel_noise = -torch.log(
        -torch.log(torch.rand_like(logits) + 1e-10) + 1e-10
    )
    perturbed_logits = (logits / temperature) + gumbel_noise * noise_scale
    return perturbed_logits


def add_exp_1_noise(
    probs: torch.Tensor, temperature: float = 1.0
) -> torch.Tensor:
    """
    Sample from unnormalized probability using the exponential race method.
    Similar to using gumbel noise, trick but we require the probs to be positive (can be unnormalized).
    You can generate samples without repeatations from the output by taking argmax or topk, etc.
    Args:
        probs: (*batch, seq_len, vocab_size) can have any number of leading dimensions.
    Returns:
        (*batch, seq_len)
    """
    exp_1_samples = 1e-10 - (torch.rand_like(probs) + 1e-10).log()
    if temperature != 1.0:
        probs = probs ** (1.0 / temperature)
    perturbed = probs / exp_1_samples
    return perturbed


def select_random_indices(
    inp_shape: torch.Size,
    num_unmask: torch.Tensor,
    select_from_mask: Optional[torch.Tensor] = None,
    selection_score: Optional[torch.Tensor] = None,
    temperature: float = 1.0,
    selection_mode: Literal["greedy", "sample"] = "greedy",
    score_mode: Literal["logits", "uprobs"] = "logits",
) -> torch.Tensor:
    """
    Select random indices from the last dimension using the selection_score.
    1. If selection score is None then it is assumed to be uniform.
    2. If score_mode = logits and selection_mode=sample, then temperature can be used to control the temperature of the distribution.
    3. If select_from_mask is provided, indices only from these positions are sampled.
    Args:
        inp_shape: torch.Size, typeically (batch, d)
        num_unmask: (batch,) int tensor
        select_from_mask: (batch, d) tensor, if provided, we only sample from the selected
        selection_score: logit-like score for selection (can be negative). Should match inp_shape, so typically (batch, d)
        score_mode:
            "logits" => p_i = \exp(s_i)/\sum_j \exp(s_j)
            "uprobs" => p_i = s_i / \sum_j s_j
    """
    if select_from_mask is not None:
        mask = select_from_mask
    else:
        # mask = torch.ones_like(inp, dtype=torch.bool)
        mask = torch.ones(
            *inp_shape, dtype=torch.bool, device=num_unmask.device
        )
    # create a tensor of shape (bs, seq) with
    # first num_unmask[i] elements of row i are 1 rest are 0
    # print(f"\nnum_unmask={num_unmask}")
    # temp = torch.nn.functional.one_hot(num_unmask, num_classes=inp.shape[-1])
    temp = torch.nn.functional.one_hot(num_unmask, num_classes=inp_shape[-1])
    # print(f"\ntemp=\n{temp}")
    temp2 = (1 - temp.cumsum(dim=-1)).to(dtype=torch.bool)
    # print(f"\ntemp2=\n{temp2.int()}")

    # now we shuffle all indices but pinning the indices for non-mask to the right end
    # rand = torch.rand(bs, seq)
    # rand = torch.rand_like(inp, dtype=torch.float32)
    if selection_score is None:  # uniform
        rand = torch.rand(
            *inp_shape, dtype=torch.float32, device=num_unmask.device
        )
    else:
        if selection_mode == "greedy":
            rand = selection_score.clone()
        elif selection_mode == "sample":
            if score_mode == "logits":
                rand = add_gumbel_noise(
                    selection_score, temperature=temperature
                )
            elif score_mode == "uprobs":
                rand = add_exp_1_noise(selection_score)
            else:
                raise ValueError(f"score_mode={score_mode} not supported")

        else:
            raise ValueError(f"selection_mode={selection_mode} not supported")
    rand[~mask] = float(
        "-inf"
    )  # to keep non-masked positions always on the right
    rand_perm = torch.argsort(rand, dim=-1, descending=True)
    # print(f"\nrand_perm=\n{rand_perm}")
    # rand_perm and temp2 have the information to index into a
    # unmask_out = torch.zeros_like(inp, dtype=torch.bool).scatter_(
    #    -1, rand_perm, temp2
    # )
    unmask_out = torch.zeros(
        *inp_shape, dtype=torch.bool, device=num_unmask.device
    ).scatter_(-1, rand_perm, temp2)
    # print(f"unmask=\n{unmask_out.int()}")
    # check if all unmask_out are mask
    assert mask[unmask_out].all()
    # check that the number of unmask matches the target
    assert (unmask_out.sum(-1) == num_unmask).all()
    return unmask_out



from typing import Dict, Any
from transformers import PreTrainedTokenizerBase


def preprocess_fn(
    example: Dict[str, Any], tokenizer: PreTrainedTokenizerBase
) -> Dict[str, Any]:
    example["token_ids"] = tokenizer.encode(  # type: ignore
        example["text"], add_special_tokens=False
    )
    return example



# %%
# change dir to the root of the project
# create the notebook inside the commands directory
import json
import os
from pathlib import Path
from typing import Any, Dict, Generator, Optional, cast

import torch

from xlm.harness import Harness
from xlm.utils.rich_utils import print_config_tree


if "PROJECT_ROOT" not in os.environ:
    os.environ["PROJECT_ROOT"] = "."
os.environ["HYDRA_FULL_ERROR"] = "1"

# region: Import necessary modules
import hydra
from omegaconf import DictConfig, OmegaConf
from lightning import seed_everything
from xlm.utils.rank_zero import RankedLogger

logger = RankedLogger(__name__, rank_zero_only=True)

# endregion


def get_generation_output_file_path(cfg: DictConfig) -> Optional[Path]:
    generation_output_dir: Optional[Path] = cfg.generation.get(
        "output_dir", None
    )
    generation_output_dir = (
        Path(generation_output_dir)
        if generation_output_dir is not None
        else None
    )
    generation_output_file_name: Optional[str] = cfg.generation.get(
        "output_file_name", None
    )
    if (
        generation_output_file_name is None
        and generation_output_dir is not None
    ):
        raise ValueError(
            "output_file_name is required when output_dir is provided"
        )
    generation_output_file_path: Optional[Path] = None
    if generation_output_file_name is not None:
        assert generation_output_dir is not None
        generation_output_file_path = (
            generation_output_dir / generation_output_file_name
        )
    return generation_output_file_path


def instantiate_model(
    cfg: DictConfig,
    datamodule: Any,
    tokenizer: Any,
) -> Harness:
    """

    Supports two modes:
        1. Load a model from full training checkpoint using `lightning_module.load_from_checkpoint(cfg.generation.ckpt_path)`
        2. Load a model from model only checkpoint using `lightning_module.model.load_state_dict(torch.load(cfg.generation.model_only_checkpoint_path))`

    Args:
        cfg: Hydra config
        generation_ckpt_path: Path to the generation checkpoint
        datamodule: Datamodule
    """
    generation_ckpt_path = cfg.generation.ckpt_path
    torch.set_float32_matmul_precision("medium")
    if generation_ckpt_path is not None:
        module_cls = hydra.utils.get_class(cfg.lightning_module._target_)
        lightning_module = module_cls.load_from_checkpoint(
            checkpoint_path=generation_ckpt_path,
            tokenizer=tokenizer,
            datamodule=datamodule,
            cfg=cfg,  # chance to override the config of the checkpoint
            manual_ema_restore=True,
        )
    else:
        lightning_module = hydra.utils.instantiate(
            cfg.lightning_module,
            tokenizer=tokenizer,
            datamodule=datamodule,
            cfg=cfg,
            _recursive_=False,
            manual_ema_restore=True,
        )
    lightning_module = cast(Harness, lightning_module)
    lightning_module = lightning_module.to("cuda")
    lightning_module.eval()

    # check if we have model only checkpoint (replicating train functionality)
    model_only_ckpt_path = None
    if cfg.generation.get("model_only_checkpoint_path", None) is not None:
        if generation_ckpt_path is not None:
            logger.error(
                "generation.model_only_checkpoint_path and generation.ckpt_path cannot both be provided. "
                "We will use generation.ckpt_path for the model weights as well."
            )
        else:
            if not os.path.isfile(cfg.model_only_checkpoint_path):
                raise ValueError(
                    f"The model only checkpoint path {cfg.model_only_checkpoint_path} does not exist."
                )
            model_only_ckpt_path = cfg.model_only_checkpoint_path

    if model_only_ckpt_path is not None:
        message = lightning_module.model.load_state_dict(
            torch.load(model_only_ckpt_path)
        )
        logger.warning(
            f"Loading weights for `model` from a pretrained model at {model_only_ckpt_path} before generation. "
            "Make sure that the model weights were saved with EMA applied."
        )
        logger.warning(message)
    return lightning_module


def generate(cfg: DictConfig):
    print_config_tree(cfg, resolve=True, save_to_file=False)
    if cfg.get("seed"):
        logger.info(f"Seed everything with seed {cfg.seed}")
        seed_everything(cfg.seed)

    # Always create the global components first.
    global_components: Dict[str, Any] = hydra.utils.instantiate(
        cfg.global_components
    )
    OmegaConf.clear_resolver("global_components")
    OmegaConf.register_new_resolver(
        "global_components", lambda x: global_components[x]
    )
    # instantiate the datamodule
    datamodule = hydra.utils.instantiate(cfg.datamodule)
    tokenizer = datamodule.tokenizer

    # update the omegaconf resolvers
    OmegaConf.clear_resolver("tokenizer")
    OmegaConf.register_new_resolver(
        "tokenizer", lambda x: getattr(tokenizer, x)
    )
    OmegaConf.clear_resolver("datamodule")
    OmegaConf.register_new_resolver(
        "datamodule", lambda x: getattr(datamodule, x)
    )
    datamodule.no_trainer_mode = True
    datamodule.prepare_data()
    datamodule.setup("predict")

    # instantiate the model
    lightning_module = instantiate_model(cfg, datamodule, tokenizer)

    # output file path
    generation_output_file_path = get_generation_output_file_path(cfg)

    if generation_output_file_path is not None:
        logger.info(
            f"Writing generation output to {generation_output_file_path}"
        )
    predict_dataloaders = datamodule.predict_dataloader()
    dataloader_names = []
    if isinstance(predict_dataloaders, list):
        for i, dl in enumerate(predict_dataloaders):
            dataloader_names.append(datamodule.dataloader_names["predict"][i])
    else:
        dataloader_names.append(datamodule.dataloader_names["predict"][0])
        predict_dataloaders = [predict_dataloaders]

    for dataloader_idx, (dataloader_name, dl) in enumerate(
        zip(dataloader_names, predict_dataloaders)
    ):
        for batch_idx, batch in enumerate(dl):
            # hardcode the move to cuda for now
            batch = {
                k: (v.to("cuda") if isinstance(v, torch.Tensor) else v)
                for k, v in batch.items()
            }
            preds = lightning_module.predict_step(
                batch, batch_idx, dataloader_idx
            )
            lightning_module._call_log_predictions(
                batch, preds, "predict", dataloader_name, no_trainer_mode=True
            )



# %%
# change dir to the root of the project
# create the notebook inside the commands directory
import os

from xlm.utils.debug import set_flags
from xlm.utils.slurm import print_slurm_info

if "PROJECT_ROOT" not in os.environ:
    os.environ["PROJECT_ROOT"] = "."
os.environ["HYDRA_FULL_ERROR"] = "1"

# region: Import necessary modules
from pathlib import Path
import hydra
from omegaconf import DictConfig, OmegaConf
from xlm.utils import omegaconf_resolvers
import dotenv

# endregion

# region: other global constants and functions
dotenv.load_dotenv(
    override=True
)  # set env variables from .env file, override=True is important
found_secretes = dotenv.load_dotenv(".secrets.env", override=True)
if not found_secretes:
    print("Warning: .secrets.env not found")
# endregion

from typing import Any, Dict, cast

import torch

from xlm.harness import Harness
from xlm.utils.rich_utils import print_config_tree
from lightning import seed_everything
from xlm.utils.rank_zero import RankedLogger
from .cli_demo import replace_model

logger = RankedLogger(__name__, rank_zero_only=True)


def instantiate_model(
    cfg: DictConfig,
    datamodule: Any,
    tokenizer: Any,
) -> Harness:
    """Instantiate a model from checkpoint or config.

    Supports two modes:
        1. Load a model from full training checkpoint using `lightning_module.load_from_checkpoint(cfg.generation.ckpt_path)`
        2. Load a model from model only checkpoint using `lightning_module.model.load_state_dict(torch.load(cfg.generation.model_only_checkpoint_path))`

    Args:
        cfg: Hydra config
        datamodule: Datamodule instance
        tokenizer: Tokenizer instance

    Returns:
        Harness: The instantiated model
    """
    generation_ckpt_path = None
    if "generation" in cfg:
        generation_ckpt_path = cfg.generation.ckpt_path
    torch.set_float32_matmul_precision("medium")
    if generation_ckpt_path is not None:
        module_cls = hydra.utils.get_class(cfg.lightning_module._target_)
        lightning_module = module_cls.load_from_checkpoint(
            checkpoint_path=generation_ckpt_path,
            tokenizer=tokenizer,
            datamodule=datamodule,
            cfg=cfg,  # chance to override the config of the checkpoint
        )
    else:
        lightning_module = hydra.utils.instantiate(
            cfg.lightning_module,
            tokenizer=tokenizer,
            datamodule=datamodule,
            cfg=cfg,
            _recursive_=False,
        )
    lightning_module = cast(Harness, lightning_module)
    lightning_module = lightning_module.to("cuda")
    lightning_module.eval()

    # check if we have model only checkpoint (replicating train functionality)
    model_only_ckpt_path = None
    if (
        "generation" in cfg
        and cfg.generation.get("model_only_checkpoint_path", None) is not None
    ):
        if generation_ckpt_path is not None:
            logger.error(
                "generation.model_only_checkpoint_path and generation.ckpt_path cannot both be provided. "
                "We will use generation.ckpt_path for the model weights as well."
            )
        else:
            if not os.path.isfile(cfg.model_only_checkpoint_path):
                raise ValueError(
                    f"The model only checkpoint path {cfg.model_only_checkpoint_path} does not exist."
                )
            model_only_ckpt_path = cfg.model_only_checkpoint_path

    if model_only_ckpt_path is not None:
        message = lightning_module.model.load_state_dict(
            torch.load(model_only_ckpt_path)
        )
        logger.warning(
            f"Loading weights for `model` from a pretrained model at {model_only_ckpt_path} before generation"
        )
        logger.warning(message)
    return lightning_module


def generate(cfg: DictConfig):
    """Generate text using the CLI demo interface."""
    print_config_tree(cfg, resolve=True, save_to_file=False)
    if cfg.get("seed"):
        logger.info(f"Seed everything with seed {cfg.seed}")
        seed_everything(cfg.seed)

    # Always create the global components first.
    global_components: Dict[str, Any] = hydra.utils.instantiate(
        cfg.global_components
    )
    OmegaConf.clear_resolver("global_components")
    OmegaConf.register_new_resolver(
        "global_components", lambda x: global_components[x]
    )
    # instantiate the datamodule
    datamodule = hydra.utils.instantiate(cfg.datamodule)
    tokenizer = datamodule.tokenizer

    # update the omegaconf resolvers
    OmegaConf.clear_resolver("tokenizer")
    OmegaConf.register_new_resolver(
        "tokenizer", lambda x: getattr(tokenizer, x)
    )
    OmegaConf.clear_resolver("datamodule")
    OmegaConf.register_new_resolver(
        "datamodule", lambda x: getattr(datamodule, x)
    )
    datamodule.no_trainer_mode = True

    # instantiate the model
    lightning_module = instantiate_model(cfg, datamodule, tokenizer)

    # get user input and predict in a loop
    with torch.inference_mode():
        while True:
            user_input = input("Enter your prompt (or 'exit' to quit): ")
            if user_input == "exit":
                break
            user_input_list = [user_input]
            preds = lightning_module.predictor.generate(user_input_list)
            print(preds[0])


# Hydra configuration parameters for CLI demo
_HYDRA_PARAMS = {
    "version_base": "1.3",
    "config_path": str(Path("../../../configs") / "lightning_train"),
    "config_name": "config.yaml",
}


@hydra.main(**_HYDRA_PARAMS)
def main(cfg: DictConfig) -> None:
    """Main function for CLI demo."""
    # global flags
    set_flags(cfg)
    # make early changes to the config
    cfg = replace_model(cfg)
    # Register resolvers
    omegaconf_resolvers.register_resolvers()
    OmegaConf.register_new_resolver(
        "datamodule", lambda attr: "${datamodule:" + str(attr) + "}"
    )
    OmegaConf.register_new_resolver(
        "tokenizer", lambda attr: "${tokenizer:" + str(attr) + "}"
    )
    OmegaConf.register_new_resolver(
        "lightning_module",
        lambda attr: "${lightning_module:" + str(attr) + "}",
    )
    OmegaConf.register_new_resolver(
        "global_components",
        lambda attr: "${global_components:" + str(attr) + "}",
    )

    # Run the generation
    generate(cfg)


if __name__ == "__main__":
    main()



# %%
# change dir to the root of the project
# create the notebook inside the commands directory
# import debugpy
#
# debugpy.listen(("localhost", 6767))
# print("Waiting for debugger attach...")
# debugpy.wait_for_client()

# fmt: off
import dotenv
# read env variables before anything else is imported
dotenv.load_dotenv(
    override=True
)  # set env variables from .env file, override=True is important
found_secretes = dotenv.load_dotenv(".secrets.env", override=True)
if not found_secretes:
    print("Warning: .secrets.env not found")
# fmt: on

import os
import json

from xlm.utils.slurm import print_slurm_info

if "PROJECT_ROOT" not in os.environ:
    os.environ["PROJECT_ROOT"] = "."
os.environ["HYDRA_FULL_ERROR"] = "1"

# region: Import necessary modules
from pathlib import Path
import hydra
from omegaconf import DictConfig, OmegaConf
from xlm.utils import omegaconf_resolvers
from xlm.external_models import setup_external_models

# endregion


# Hydra configuration parameters
# Use absolute path to configs directory within the package
_HYDRA_PARAMS = {
    "version_base": "1.3",
    "config_path": str(
        (
            Path(__file__).parent.parent / "configs" / "lightning_train"
        ).resolve()
    ),
    "config_name": "config.yaml",
}

# Register a temporary resolver so that early calls to resolve() don't fail
omegaconf_resolvers.register_resolvers()
OmegaConf.register_new_resolver(
    "datamodule", lambda attr: "${datamodule:" + str(attr) + "}"
)
OmegaConf.register_new_resolver(
    "tokenizer", lambda attr: "${tokenizer:" + str(attr) + "}"
)
OmegaConf.register_new_resolver(
    "lightning_module", lambda attr: "${lightning_module:" + str(attr) + "}"
)
OmegaConf.register_new_resolver(
    "global_components", lambda attr: "${global_components:" + str(attr) + "}"
)
# endregion

# Setup external models before Hydra initialization
# This adds external model directories to sys.path for imports

from hydra.core.plugins import Plugins
from hydra.plugins.search_path_plugin import SearchPathPlugin
from hydra.core.config_search_path import ConfigSearchPath

hydra_plugins = Plugins.instance()
class HydraSearchPathPlugin(SearchPathPlugin):
        def manipulate_search_path(
            self, search_path: ConfigSearchPath
        ) -> None:
            search_path.append("file", str(Path(__file__).parent.parent / "configs/common"))

hydra_plugins.register(HydraSearchPathPlugin)

external_model_dirs = setup_external_models()
# Register our SearchPathPlugin manually with Hydra
if external_model_dirs:
    class ExternalModelsSearchPathPlugin(SearchPathPlugin):
        def manipulate_search_path(
            self, search_path: ConfigSearchPath
        ) -> None:
            for model_dir in external_model_dirs:
                config_dir = model_dir / "configs"
                if config_dir.exists():
                    search_path.append("file", str(config_dir))

    # Register the plugin
    hydra_plugins.register(ExternalModelsSearchPathPlugin)


@hydra.main(**_HYDRA_PARAMS)
def main(cfg: DictConfig) -> None:
    # Print predictor config as JSON and exit if requested
    if getattr(cfg, "job_type", None) == "print_predictor_params":
        out_path = getattr(cfg, "predictor_json_out", None)
        data = json.dumps(
            OmegaConf.to_container(cfg.get("predictor", {}), resolve=True)
        )
        if out_path:
            with open(out_path, "w") as f:
                f.write(data)
        else:
            print(data)
        return
    from xlm.utils.rich_utils import print_config_tree

    if cfg.job_type == "name":
        job_name = cfg.job_name

        print_config_tree(cfg, resolve=True, save_to_file=False)
        print(f"job_name: \t{job_name}")
        return

    if cfg.job_type == "length_analysis":
        raise NotImplementedError("Length analysis is not implemented yet")

    from xlm.utils.debug import set_flags
    from xlm.commands.lightning_train import train
    from xlm.commands.lightning_eval import evaluate
    from xlm.commands.lightning_generate import generate
    from xlm.commands.lightning_prepare_data import prepare_data

    if cfg.job_type == "train":
        import multiprocessing as mp

        # at some point I started getting worker error
        # I tried the workaround mentioned here: https://github.com/pytorch/pytorch/issues/119845 buy setting mp.set_start_method("spawn")
        # if the issue reappears, try uncommenting the line below
        # mp.set_start_method("spawn")
        print_config_tree(cfg, resolve=True, save_to_file=cfg.paths.run_dir)
        set_flags(cfg)
        print_slurm_info()
        train(cfg)
    elif cfg.job_type == "eval":
        set_flags(cfg)
        evaluate(cfg)
    elif cfg.job_type == "generate":
        generate(cfg)
    elif cfg.job_type == "prepare_data":
        print_config_tree(cfg, resolve=True)
        prepare_data(cfg)


if __name__ == "__main__":
    main()



# v2
from typing import Protocol, Tuple
from torch import Tensor as TT
from jaxtyping import Float


class NoiseSchedule(Protocol):
    def noise_rate(self, t: Float[TT, " *batch"]) -> Float[TT, " *batch"]: ...

    def total_noise(self, t: Float[TT, " *batch"]) -> Float[TT, " *batch"]: ...

    def t_from_noise_rate(self, noise_rate: float) -> float: ...

    def t_from_total_noise(self, total_noise: float) -> float: ...

    def sample_t(self, batch_size: int) -> Float[TT, " batch_size"]: ...

    def forward(
        self, t: Float[TT, " *batch"]
    ) -> Tuple[Float[TT, " *batch"], Float[TT, " *batch"]]: ...

    def __call__(
        self, t: Float[TT, " *batch"]
    ) -> Tuple[Float[TT, " *batch"], Float[TT, " *batch"]]: ...


class DummyNoiseSchedule(NoiseSchedule):
    def noise_rate(self, t: Float[TT, "batch"]) -> Float[TT, "batch"]:
        raise NotImplementedError("This noise schedule does nothing.")

    def total_noise(self, t: Float[TT, "batch"]) -> Float[TT, "batch"]:
        raise NotImplementedError("This noise schedule does nothing.")

    def t_from_noise_rate(self, noise_rate: float) -> float:
        raise NotImplementedError("This noise schedule does nothing.")

    def t_from_total_noise(self, total_noise: float) -> float:
        raise NotImplementedError("This noise schedule does nothing.")

    def sample_t(self, batch_size: int) -> Float[TT, "batch_size"]:
        raise NotImplementedError("This noise schedule does nothing.")

    def forward(
        self, t: Float[TT, " *batch"]
    ) -> Tuple[Float[TT, " *batch"], Float[TT, " *batch"]]:
        raise NotImplementedError("This noise schedule does nothing.")

    def __call__(
        self, t: Float[TT, " *batch"]
    ) -> Tuple[Float[TT, " *batch"], Float[TT, " *batch"]]:
        raise NotImplementedError("This noise schedule does nothing.")
